from image_fetcher.get_image_urls import get_image_urls
from image_fetcher.validate_params import validate_download_images_params
from image_fetcher.download_page import download_page
from image_fetcher.download_image import download_image
from image_fetcher.tools import get_existing_images, print_summary
from image_fetcher.browsers import Browser, BrowserType

from tqdm import tqdm


def download_images(search_term, total_images, headers, browser, 
extensions=['jpg','png'], directory=None, progress_bar=True, verbose=True):
    """
    Downloads images from google for given search_term

    Parameters:
    search_term (str): Given term for downloading images from google
    total_images (int): Total number of images to download (will skip identical images in the given directory)
    headers (dict): Headers for urllib to use when requesting from urls (must include 'User-Agent')
    extensions (list): Image file extensions to accept (defaults to 'jpg' and 'png')
    directory (str): Directory to save images to (defaults to naming folder same as search_term)
    progress_bar (bool): Whether to display progress bar during downloading (defaults to True)
    verbose (bool): Whether to print total downloaded & total ignored at the end (defaults to True)
    """
    if not directory:
        directory = search_term
    #Validate passed params
    validate_download_images_params(search_term, total_images, extensions, headers, browser, directory, verbose, progress_bar)
    #Setup variables
    #Download raw HTML from google image search of given term
    page = download_page(search_term, total_images, browser)
    #Get list of iamge URLS from the page
    urls = get_image_urls(page, verbose=verbose)

    existing_images = get_existing_images(directory)
    images_in_folder = len(existing_images)
    total_downloaded = 0
    #If progress bar initialise tqdm
    if progress_bar:
        pbar = tqdm(total=total_images)
        pbar.update(images_in_folder)
    
    #Download urls from url list
    url_index = 0
    while total_downloaded+images_in_folder < total_images:
        #Try to download next URL
        image = download_image(urls[url_index], directory, headers, existing_images, extensions)
        url_index += 1
        
        if image:
            #if image was downloaded
            if image == 1:
                total_downloaded+=1
            if progress_bar:
                pbar.update(1)

        #If we've run out of URL's due to them being erroneous we'll get more
        if url_index == len(urls):
            if verbose:
                print("All URL's attempted, fetching more")
            page = download_page(search_term, total_images+100, browser)
            urls = get_image_urls(page, verbose=verbose)

    if progress_bar:
        pbar.close()
    if verbose:
        print_summary(search_term, total_downloaded=total_downloaded, total_ignored=images_in_folder)



from image_fetcher.browsers import Browser

from os import listdir

def is_alphanumeric(string):
    return all(char.isalnum() or char.isspace() for char in string)

def validate_directory(directory):
    if not isinstance(directory, str):
        raise TypeError("directory must be a string, "+directory+" is not")
    if not is_alphanumeric(directory):
        raise ValueError("directory must be alphanumeric ('/' added automatically)")

def validate_extensions(extensions):
    if not isinstance(extensions, list):
        raise TypeError("extensions must be a list")
    for extension in extensions:
        if not extension.isalpha():
            raise ValueError("extensions must be alpha, "+extension+" is not")

def validate_headers(headers):
    if not isinstance(headers, dict):
        raise TypeError("headers must be a dict, "+headers+" is not")
    if 'User-Agent' not in headers:
        raise ValueError("headers must contain User-Agent")

def validate_positive_number(var, name):
    if not isinstance(var, int):
        raise TypeError(name+" must be an integer, "+var+" is not")
    if var < 1:
        raise ValueError(name+" must be greater than 0")

def validate_search_term(search_term):
    if not isinstance(search_term, str):
        raise ValueError("search_term must be a string")
    if not is_alphanumeric(search_term):
        raise ValueError("search_term must be alphanumeric")

def validate_bool(var, name):
    if not isinstance(var, bool):
        raise TypeError(name+" must be a bool")

def validate_browser(browser):
    if not browser or not isinstance(browser, Browser):
        raise TypeError("browser must be a image_fetcher.browsers.Browser")

def validate_image_fetching_arguments(total_images, extensions, headers, verbose, progress_bar):
    validate_positive_number(total_images, "total_images")
    validate_extensions(extensions)
    validate_headers(headers)
    validate_bool(verbose,"verbose")
    validate_bool(progress_bar,"progress_bar")

def validate_download_images_params(search_term, total_images, extensions, headers, browser, directory, verbose, progress_bar):
    validate_browser(browser)
    validate_search_term(search_term)
    validate_directory(directory)
    validate_image_fetching_arguments(total_images, extensions, headers, verbose, progress_bar)

def validate_download_image_params(url, directory, headers, existing_images, extensions, raise_errors):
    if not isinstance(url, str):
        raise ValueError("url must be a string")
    validate_directory(directory)
    validate_headers(headers)
    if existing_images and not isinstance(existing_images, list):
        raise TypeError("existing_images must be a list")
    validate_extensions(extensions)
    validate_bool(raise_errors,"raise_errors")

def validate_concurrent_images_download(search_term, total_images, headers, browser, max_image_fetching_threads, extensions, directory, progress_bar, verbose):
    validate_browser(browser)
    validate_search_term(search_term)
    validate_directory(directory)
    validate_positive_number(max_image_fetching_threads, "max_image_fetching_threads")
    validate_image_fetching_arguments(total_images, extensions, headers, verbose, progress_bar)
    
def validate_concurrent_image_search_params(search_terms, total_images, headers, browser, max_similtanous_threads, max_image_fetching_threads, extensions, directories, progress_bar, verbose):
    validate_browser(browser)
    if not isinstance(search_terms, list):
        raise TypeError("search_terms must be a list")
    for search_term in search_terms:
        validate_search_term(search_term)
    if not isinstance(directories, list):
        raise TypeError("search_terms must be a list")
    for directory in directories:
        validate_directory(directory)
    if len(directories) != len(search_terms):
        raise ValueError("directories and search_terms must match in length")
    validate_positive_number(max_similtanous_threads, "max_similtanous_threads")
    validate_positive_number(max_image_fetching_threads, "max_image_fetching_threads")
    validate_image_fetching_arguments(total_images, extensions, headers, verbose, progress_bar)



from urllib import request
from json import loads


def download_url(url, headers):
    """
    Read data from a given URL

    Parameters:
    url (str): URL to retrieve data from
    headers (dict): Headers for urllib to use when requesting from urls (must include 'User-Agent')

    Returns:
    bytes: Returns bytes of data from the URL
    """
    req = request.Request(url, headers=headers)
    resp = request.urlopen(req)
    data = resp.read()
    resp.close()
    return data


def get_image_urls(page, verbose=True):
    """
    Get a list of URL's from a given google search page

    Parameters:
    page (str): HTML source of a google image search
    verbose (bool): Whether to print total URL's found (defaults to True)

    Returns:
    list: Returns list of image URL's
    """
    image_urls = []
    #Get all image source URLS
    for image_html in page.find_all("div",{"class":"rg_meta"}):
        try:
            #Add the image source to urls
            final_image_url = loads(image_html.text)["ou"]
            image_urls.append(final_image_url)
        except:
            break
    if verbose:
        print(str(len(image_urls))+" image urls found")
    return image_urls



from image_fetcher.tools import round_up_to_nearest_hundred
from image_fetcher.browsers import BrowserType

from selenium import webdriver
from bs4 import BeautifulSoup
from time import sleep


wait_time = 1
#As some URL's will likely fail, It's a good idea to fetch more than we need
#The backup_threshold is what amount to multiple needed URL's by
#At a rating of 2, for 200 images we'd fetch 400 URL's (leaving 200 to spare)
backup_threshold = 2
google_images = 'https://www.google.com/search?tbm=isch&q='

options = webdriver.ChromeOptions()
options.add_argument('--no-sandbox')
options.add_argument('--headless')

def launch_driver(browser):
    if browser.browser_type == BrowserType.CHROME:
        return webdriver.Chrome(browser.driver, options=browser.options)
    elif browser.browser_type == BrowserType.FIREFOX:
        return webdriver.Firefox(executable_path=browser.driver, options=browser.options)

def download_page(search_term, total_images, browser):
    driver = launch_driver(browser)
    driver.get(google_images+search_term)
    total_images *= backup_threshold
    required_scrolls = int(round_up_to_nearest_hundred(total_images) / 100)-1
    for i in range(0, required_scrolls):
        driver.execute_script('window.scrollTo(0,document.body.scrollHeight);')
        sleep(wait_time)
    source = driver.page_source
    driver.close()
    return BeautifulSoup(source, 'html.parser')



from selenium import webdriver
from enum import Enum
from os.path import abspath
from os import listdir


class BrowserType(Enum):
    CHROME = 1
    FIREFOX = 2


def validate_driver(driver_path):
    if "/" in driver_path:
        driver = driver_path.split("/")[-1]
        directory = driver_path[0:len(driver_path)-len(driver)-1]
        if not driver in listdir(directory):
            raise ValueError(driver+" not found in "+directory)
    else:
        driver = driver_path
        if not driver in listdir():
            raise ValueError(driver+" not found in current directory")


class Browser:
    def __init__(self, browser_type, driver):
        validate_driver(driver)
        self.browser_type = browser_type
        if browser_type == BrowserType.CHROME:
            chrome_options = webdriver.ChromeOptions()
            chrome_options.add_argument('--no-sandbox')
            chrome_options.add_argument('--headless')
            self.driver = driver
            self.options = chrome_options
        elif browser_type == BrowserType.FIREFOX:
            firefox_options = webdriver.firefox.options.Options()
            firefox_options.headless = True
            self.driver = abspath(driver)
            self.options = firefox_options
        else:
            raise ValueError("driver_type not valid, choose from: "+str(list(Browser)))



from image_fetcher.get_image_urls import download_url
from image_fetcher.tools import get_extension, escape_image_name
from image_fetcher.validate_params import validate_download_image_params
from image_fetcher.download_page import download_page

from tqdm import tqdm
from os import listdir
from func_timeout import func_timeout, FunctionTimedOut


def download_image_simple_with_timeout(url, timeout, directory, headers, existing_images=[], extensions=['jpg','png']):
    """
    Downloads image from given URL. Will exit if not complete within timeout seconds. Doesn't validate input params.

    Parameters:
    url (str): URL to try and download image from
    timeout (int): Seconds to wait for function to execute
    directory (str): directory to save image to
    existing_images (list): list of existing images to avoid trying to download an existing image 
    (can be fetched by passing the folder name to get_existing_images(), default is []) 
    extensions (list): Image file extensions to accept (defaults to 'jpg' and 'png')
    """
    try:
        func_timeout(timeout, download_image_simple, args=(url, directory, headers, existing_images, extensions,))
    except FunctionTimedOut:
        pass


def download_image_simple(url, directory, headers, existing_images=[], extensions=['jpg','png']):
    """
    Downloads image from given URL. Doesn't validate input params.

    Parameters:
    url (str): URL to try and download image from
    directory (str): directory to save image to
    existing_images (list): list of existing images to avoid trying to download an existing image 
    (can be fetched by passing the folder name to get_existing_images(), default is []) 
    extensions (list): Image file extensions to accept (defaults to 'jpg' and 'png')
    """
    try:
        if get_extension(url) in extensions:
            image_name = escape_image_name(url)+'.'+get_extension(url)
            if image_name not in existing_images:
                data = download_url(url, headers)
                output_file = open(directory+'/'+image_name, 'wb')
                output_file.write(data)
                output_file.close()
    except:
        pass


def download_image(url, directory, headers, existing_images=[], extensions=['jpg','png'], raise_errors=False):
    """
    Downloads image from given URL and returns status code

    Parameters:
    url (str): URL to try and download image from
    directory (str): directory to save image to
    existing_images (list): list of existing images to avoid trying to download an existing image 
    (can be fetched by passing the folder name to get_existing_images(), default is []) 
    extensions (list): Image file extensions to accept (defaults to 'jpg' and 'png')
    raise_errors (bool): Whether to raise an error if one occurs during download (default is False)

    Returns:
    int: Returns 1 if an image was downloaded, 2 if the download was ignored as the image already existed, 
    or 0 if it failed/was aborted due to invalid file type
    """
    validate_download_image_params(url, directory, headers, existing_images, extensions, raise_errors)
    try:
        if get_extension(url) in extensions:
            #Convert escaped URL to the image name (used to ensure same image isn't downloaded again in future)
            image_name = escape_image_name(url)+'.'+get_extension(url)
            if image_name not in existing_images:
                data = download_url(url, headers)
                #Write data from URL request to image
                output_file = open(directory+'/'+image_name, 'wb')
                output_file.write(data)
                output_file.close()
                if image_name in listdir(directory):
                    return 1
            else:
                return 2
    except:
        if raise_errors:
            raise 
        pass
    return 0



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="image_fetcher",
    version="1.1.3",
    author="Ben Andrew",
    author_email="benandrew89@gmail.com",
    description="Package for downloading images from google",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BenAAndrew/Python-Image-Fetcher",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "beautifulsoup4>=4.8.0",
        "bs4>=0.0.1",
        "certifi>=2019.9.11",
        "chardet>=3.0.4",
        "func-timeout>=4.3.5",
        "idna>=2.8",
        "selenium>=3.141.0",
        "soupsieve>=1.9.3",
        "tqdm>=4.36.1",
        "urllib3>=1.25.5",
        "wincertstore>=0.2",
    ],
    python_requires='>=3.6',
)


from image_fetcher.validate_params import validate_directory

from os import listdir, mkdir
from math import ceil
from re import sub


def get_existing_images(directory):
    """
    Get's a list of files in a given directory or creates it if the directory doesn't exist

    Parameters:
    directory (str): directory to search

    Returns:
    list: Returns list of files in the directory (expty if the directory didn't exist)
    """
    validate_directory(directory)
    directory += '/'
    try:
        return listdir(directory)
    except:
        mkdir(directory)
        return []


def get_extension(url):
    return url.split(".")[-1]


def escape_image_name(url):
    """
    Get URL without extension and 'http(s)://' and removes all non alphanumeric characters

    Parameters:
    url (str): URL to escape

    Returns:
    str: Escaped URL
    """
    #Remove extension
    escaped = url[:len(url)-len(url.split('.')[1])-1]
    #Remove http:// or https://
    escaped = escaped.split('/',1)[1]
    #Remove all non alphanumeric characters
    return sub("[^a-zA-Z0-9]+", '', escaped)


def round_up_to_nearest_hundred(x):
    return int(ceil(x / 100.0)) * 100


def print_summary(search_term, total_downloaded, total_ignored):
    print(search_term+" DONE")
    print("Total downloaded = "+str(total_downloaded))
    print("Total ignored as they already existed = "+str(total_ignored))



from image_fetcher.get_image_urls import get_image_urls
from image_fetcher.validate_params import validate_concurrent_images_download, validate_concurrent_image_search_params
from image_fetcher.download_images import get_existing_images
from image_fetcher.download_image import download_image_simple_with_timeout
from image_fetcher.download_page import download_page
from image_fetcher.tools import print_summary

from tqdm import tqdm
from os import listdir
from concurrent.futures import ThreadPoolExecutor, wait


def concurrent_images_download(search_term, total_images, headers, max_image_fetching_threads, image_download_timeout, 
browser, extensions=['jpg','png'], directory=None, progress_bar=True, verbose=True):
    """
    Downloads images from google for given search_term using multiple concurrent threads

    Parameters:
    search_term (str): Given term for downloading images from google
    total_images (int): Total number of images to download (will skip identical images in the given directory)
    headers (dict): Headers for urllib to use when requesting from urls (must include 'User-Agent')
    max_image_fetching_threads (int): Maximum number of similtaneously running image fetching threads
    extensions (list): Image file extensions to accept (defaults to 'jpg' and 'png')
    directory (str): Directory to save images to (defaults to naming folder same as search_term)
    progress_bar (bool): Whether to display progress bar during downloading (defaults to True)
    verbose (bool): Whether to print total downloaded & total ignored at the end (defaults to True)
    """
    if not directory:
        directory = search_term
    #Validate passed params
    validate_concurrent_images_download(search_term, total_images, headers, browser, max_image_fetching_threads, extensions, directory, progress_bar, verbose)
    #Setup variables
    #Download raw HTML from google image search of given term
    page = download_page(search_term, total_images, browser)
    #Get list of iamge URLS from the page
    urls = get_image_urls(page, verbose=verbose)

    #Setup thread executor pool and active threads list
    pool = ThreadPoolExecutor(max_image_fetching_threads)
    futures = []

    #Get existing images 
    existing_images = get_existing_images(directory)
    images_in_folder = len(existing_images)
    
    #If progress bar initialise tqdm
    if progress_bar:
        pbar = tqdm(total=total_images)
        pbar.update(images_in_folder)
    
    url_index = 0 
    while images_in_folder != total_images:
        #Append maximum required number of threads (pool will limit the number of ones running concurrently)
        for x in range(0, total_images-images_in_folder):
            futures.append(pool.submit(download_image_simple_with_timeout, urls[url_index], image_download_timeout, directory, headers, existing_images, extensions))
            #Increment url_index so each call to download_image will take a different url
            url_index+=1

        #Wait for all threads to execute
        wait(futures)

        if progress_bar:
            images_in_folder_before = images_in_folder
        #Count new total of images in the folder to either exit the loop or execute more threads
        images_in_folder = len(listdir(directory))
        if progress_bar:
            pbar.update(images_in_folder-images_in_folder_before)

        #If we've run out of URL's due to them being erroneous we'll get more
        if url_index+(total_images-images_in_folder) >= len(urls):
            if verbose:
                print("All URL's attempted, fetching more")
            page = download_page(search_term, total_images+100, browser.webdriver)
            urls = get_image_urls(page, verbose=verbose)

    if progress_bar:
        pbar.close()
    if verbose:
        print_summary(search_term, total_downloaded=total_images-len(existing_images), total_ignored=len(existing_images))


def concurrent_image_search(search_terms, total_images, headers, max_similtanous_threads, max_image_fetching_threads, 
image_download_timeout, browser, extensions=['jpg','png'], directories=None, progress_bar=True, verbose=True):
    """
    Downloads images from google for multiple search_terms using multiple concurrent threads

    Parameters:
    search_terms (list): List of search terms for downloading images from google
    total_images (int): Total number of images to download (will skip identical images in the given directory)
    headers (dict): Headers for urllib to use when requesting from urls (must include 'User-Agent')
    max_similtanous_threads (int): Maximum number of similtaneously running image search threads
    max_image_fetching_threads (int): Maximum number of similtaneously running image fetching downloading per image search
    extensions (list): Image file extensions to accept (defaults to 'jpg' and 'png')
    directories (list): Directories to save images to (defaults to naming folder same as search_term)
    progress_bar (bool): Whether to display progress bar during downloading (defaults to True)
    verbose (bool): Whether to print total downloaded & total ignored at the end (defaults to True)
    """
    if not directories:
        directories = search_terms
    #Validate passed params
    validate_concurrent_image_search_params(search_terms, total_images, headers, browser, max_similtanous_threads, max_image_fetching_threads, extensions, directories, progress_bar, verbose)
    #Setup thread executor pool and active threads list
    pool = ThreadPoolExecutor(max_image_fetching_threads)
    futures = []
    for i in range(0, len(search_terms)):
        #Add new concurrent_images_download process thread for each search term
        futures.append(pool.submit(concurrent_images_download, search_terms[i], total_images, headers, max_image_fetching_threads, image_download_timeout, browser, extensions, directories[i], progress_bar, verbose))
    #Wait for all threads to execute
    wait(futures)
